/*
 *Copyright (C) 1994-2000 The XFree86 Project, Inc. All Rights Reserved.
 *
 *Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 *"Software"), to deal in the Software without restriction, including
 *without limitation the rights to use, copy, modify, merge, publish,
 *distribute, sublicense, and/or sell copies of the Software, and to
 *permit persons to whom the Software is furnished to do so, subject to
 *the following conditions:
 *
 *The above copyright notice and this permission notice shall be
 *included in all copies or substantial portions of the Software.
 *
 *THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *NONINFRINGEMENT. IN NO EVENT SHALL THE XFREE86 PROJECT BE LIABLE FOR
 *ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 *CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *Except as contained in this notice, the name of the XFree86 Project
 *shall not be used in advertising or otherwise to promote the sale, use
 *or other dealings in this Software without prior written authorization
 *from the XFree86 Project.
 *
 * Authors:	Harold L Hunt II
 *		MATSUZAKI Kensuke
 */
/* $XFree86: xc/programs/Xserver/hw/xwin/winwindow.c,v 1.3 2002/10/17 08:18:25 alanh Exp $ */

#include "win.h"

/*
 * Prototypes for local functions
 */

static int
winAddRgn (WindowPtr pWindow, pointer data);

static void
winUpdateRgn (WindowPtr pWindow);

#ifdef SHAPE
static void
winReshape (WindowPtr pWin);
#endif

static void
winCreateWindowsWindow(WindowPtr pWin);

static void
winDestroyWindowsWindow(WindowPtr pWin);

static void
winUpdateWindowsWindow(WindowPtr pWin);

static XID
winGetWindowID(WindowPtr pWin);


static BOOL fMoveByXServer = FALSE;//??????
static UINT s_nIDMoveEvent = 0;

#define MOVE_EVENT_ELAPSE 200

/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbCreateWindow() */

Bool
winCreateWindowNativeGDI (WindowPtr pWin)
{
  ErrorF ("winCreateWindow()\n");
  return TRUE;
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbDestroyWindow() */

Bool
winDestroyWindowNativeGDI (WindowPtr pWin)
{
  ErrorF ("winDestroyWindow()\n");
  return TRUE;
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbPositionWindow() */

Bool
winPositionWindowNativeGDI (WindowPtr pWin, int x, int y)
{
  ErrorF ("winPositionWindow()\n");
  return TRUE;
}


/* See Porting Layer Definition - p. 39 */
/* See mfb/mfbwindow.c - mfbCopyWindow() */

void 
winCopyWindowNativeGDI (WindowPtr pWin,
			DDXPointRec ptOldOrg,
			RegionPtr prgnSrc)
{
  ErrorF ("winCopyWindow()\n");
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbChangeWindowAttributes() */

Bool
winChangeWindowAttributesNativeGDI (WindowPtr pWin, unsigned long mask)
{
  ErrorF ("winChangeWindowAttributes()\n");
  return TRUE;
}


/* See Porting Layer Definition - p. 37
 * Also referred to as UnrealizeWindow
 */

Bool
winUnmapWindowNativeGDI (WindowPtr pWindow)
{
  ErrorF ("winUnmapWindow()\n");
  /* This functions is empty in the CFB,
   * we probably won't need to do anything
   */
  return TRUE;
}


/* See Porting Layer Definition - p. 37
 * Also referred to as RealizeWindow
 */

Bool
winMapWindowNativeGDI (WindowPtr pWindow)
{
  ErrorF ("winMapWindow()\n");
  /* This function is empty in the CFB,
   * we probably won't need to do anything
   */
  return TRUE;

}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbCreateWindow() */

Bool
winCreateWindowPRootless (WindowPtr pWin)
{
  Bool			fResult = FALSE;
  winWindowPriv(pWin);

#if CYGDEBUG
  ErrorF ("winCreateWindowPRootless()\n");
#endif
  ErrorF ("%08X:winCreateWindowPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->CreateWindow(pWin);
  
  pWinPriv->hRgn = NULL;
  pWinPriv->hWnd = NULL;
  pWinPriv->pScreenPriv = winGetScreenPriv(pWin->drawable.pScreen);
  pWinPriv->fXKilled = FALSE;
  
  /*winUpdateRgn (pWin);*/

  return fResult;
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbDestroyWindow() */

Bool
winDestroyWindowPRootless (WindowPtr pWin)
{
  Bool			fResult = FALSE;
  winWindowPriv(pWin);

#if CYGDEBUG
  ErrorF ("winDestroyWindowPRootless()\n");
#endif
  ErrorF ("%08X:winDestroyWindowPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->DestroyWindow(pWin);
  
  pWinPriv->fXKilled = TRUE;
  
  if (pWinPriv->hRgn != NULL)
    {
      DeleteObject(pWinPriv->hRgn);
      pWinPriv->hRgn = NULL;
    }
  
  winUpdateRgn (pWin);
  winDestroyWindowsWindow (pWin); 

  return fResult;
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbPositionWindow() */

Bool
winPositionWindowPRootless (WindowPtr pWin, int x, int y)
{
  Bool			fResult = FALSE;
  int		        iX, iY, iWidth, iHeight, iBorder;
  winWindowPriv(pWin);
  HWND hWnd = pWinPriv->hWnd;
  RECT rcNew;
  RECT rcOld;
  RECT rcClient;
  RECT *lpRc;
  DWORD dwExStyle;
  DWORD dwStyle;

#if CYGDEBUG
  ErrorF ("winPositionWindowPRootless()\n");
#endif
  ErrorF ("%08X:winPositionWindowPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->PositionWindow(pWin, x, y);
  
  if (wBoundingShape(pWin))
    {
      winReshape (pWin);
    }
  
  winUpdateRgn (pWin);
  
  if (1)
  if (hWnd)
    {
      dwExStyle = GetWindowLongPtr (hWnd,
				    GWL_EXSTYLE);
      dwStyle = GetWindowLongPtr (hWnd,
				    GWL_STYLE);

      iBorder = wBorderWidth (pWin);
      
      iX = pWin->drawable.x - iBorder;
      iY = pWin->drawable.y - iBorder;
      
      iWidth = pWin->drawable.width;// + iBorder * 2;
      iHeight = pWin->drawable.height;// + iBorder * 2;
      
      SetRect (&rcNew, iX, iY, iX + iWidth, iY + iHeight);
      
      lpRc = &rcNew;
      ErrorF ("(%d ms)drawable (%d, %d)-(%d, %d)\n", GetTickCount(), lpRc->left, lpRc->top, lpRc->right, lpRc->bottom);
      
      AdjustWindowRectEx (&rcNew, dwStyle, FALSE, dwExStyle);
      
      GetWindowRect (hWnd, &rcOld);
      GetClientRect (hWnd, &rcClient);
      
      lpRc = &rcNew;
      ErrorF ("(%d ms)rcNew (%d, %d)-(%d, %d)\n", GetTickCount(), lpRc->left, lpRc->top, lpRc->right, lpRc->bottom);
      
      lpRc = &rcOld;
      ErrorF ("(%d ms)rcOld (%d, %d)-(%d, %d)\n", GetTickCount(), lpRc->left, lpRc->top, lpRc->right, lpRc->bottom);
      
      lpRc = &rcClient;
      ErrorF ("(%d ms)rcClient (%d, %d)-(%d, %d)\n", GetTickCount(), lpRc->left, lpRc->top, lpRc->right, lpRc->bottom);
      
      if (!EqualRect (&rcNew, &rcOld))
	{
	  ErrorF("need to move\n");
	  //fMoveByXServer = TRUE;
	  MoveWindow (hWnd, rcNew.left, rcNew.top,
		      rcNew.right - rcNew.left, rcNew.bottom - rcNew.top, TRUE);
	  fMoveByXServer = FALSE;
	}
      else
	{
	    ErrorF("not need to move\n");
	}
      //Sleep (5000);
    }
  return fResult;
}


/* See Porting Layer Definition - p. 37 */
/* See mfb/mfbwindow.c - mfbChangeWindowAttributes() */

Bool
winChangeWindowAttributesPRootless (WindowPtr pWin, unsigned long mask)
{
  Bool			fResult = FALSE;

#if CYGDEBUG
  ErrorF ("winChangeWindowAttributesPRootless()\n");
#endif
  ErrorF ("%08X:winChangeWindowAttributesPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->ChangeWindowAttributes(pWin, mask);
  
  winUpdateRgn (pWin);
  
  return fResult;
}


/* See Porting Layer Definition - p. 37
 * Also referred to as UnrealizeWindow
 */

Bool
winUnmapWindowPRootless (WindowPtr pWin)
{
  Bool			fResult = FALSE;
  winWindowPriv(pWin);

#if CYGDEBUG
  ErrorF ("winUnmapWindowPRootless()\n");
#endif
  ErrorF ("%08X:winUnmapWindowPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->UnrealizeWindow(pWin);
  
  pWinPriv->fXKilled = TRUE;
  
  winUpdateRgn (pWin);
 
  winDestroyWindowsWindow (pWin);
  return fResult;
}


/* See Porting Layer Definition - p. 37
 * Also referred to as RealizeWindow
 */

Bool
winMapWindowPRootless (WindowPtr pWin)
{
  Bool			fResult = FALSE;
  winWindowPriv(pWin);

#if CYGDEBUG
  ErrorF ("winMapWindowPRootless()\n");
#endif
  ErrorF ("%08X:winMapWindowPRootless\n", pWin);

  fResult = winGetScreenPriv(pWin->drawable.pScreen)->RealizeWindow(pWin);
  
  pWinPriv->fXKilled = FALSE;
  
  winUpdateRgn (pWin);
  
  winUpdateWindowsWindow (pWin);
  
  ShowWindow (pWinPriv->hWnd, SW_SHOWNORMAL);
  
  /* Send first paint message */
  UpdateWindow (pWinPriv->hWnd);
  
  return fResult;
}

void
winReparentPRootless (WindowPtr pWin, WindowPtr pPriorParent)
{
  ErrorF ("%08X:winReparentPRootless\n", pWin);
  winUpdateWindowsWindow (pWin);
}

#ifdef SHAPE
void
winSetShapePRootless (WindowPtr pWin)
{
#if CYGDEBUG
  ErrorF ("winSetShapePRootless()\n");
#endif
  ErrorF ("%08X:winSetShapePRootless\n", pWin);

  winGetScreenPriv(pWin->drawable.pScreen)->SetShape(pWin);
  
  winReshape (pWin);
  winUpdateRgn (pWin);
  
  return;
}
#endif


/*
 * Local function for adding a region to the Windows window region
 */

static
int
winAddRgn (WindowPtr pWin, pointer data)
{
  int		iX, iY, iWidth, iHeight, iBorder;
  HRGN		hRgn = *(HRGN*)data;
  HRGN		hRgnWin;
  winWindowPriv(pWin);
  
  /* If pWin is not Root */
  if (pWin->parent != NULL) 
    {
#if CYGDEBUG
      ErrorF("winAddRgn()\n");
#endif
      if (pWin->mapped)
	{
	  iBorder = wBorderWidth (pWin);
	  
	  iX = pWin->drawable.x;// - iBorder;
	  iY = pWin->drawable.y;// - iBorder;
	  
	  iWidth = pWin->drawable.width;// + iBorder * 2;
	  iHeight = pWin->drawable.height;// + iBorder * 2;
	  
	  hRgnWin = CreateRectRgn (0, 0, iWidth, iHeight);
	  
	  if (hRgnWin == NULL)
	    {
	      ErrorF ("winAddRgn - CreateRectRgn() failed\n");
	      ErrorF ("  Rect %d %d %d %d\n",
		      iX, iY, iX + iWidth, iY + iHeight);
	    }
	  
	  if (pWinPriv->hRgn)
	    {
	      if (CombineRgn (hRgnWin, hRgnWin, pWinPriv->hRgn, RGN_AND)
		  == ERROR)
		{
		  ErrorF ("winSetShapePRootless - CombineRgn() failed\n");
		}
	    }
	  
	  OffsetRgn (hRgnWin, iX, iY);

	  if (CombineRgn (hRgn, hRgn, hRgnWin, RGN_OR) == ERROR)
	    {
	      ErrorF ("winAddRgn - CombineRgn() failed\n");
	    }
	  
	  DeleteObject (hRgnWin);
	}
      return WT_DONTWALKCHILDREN;
    }
  else
    {
      return WT_WALKCHILDREN;
    }
}


/*
 * Local function to update the Windows window's region
 */

static
void
winUpdateRgn (WindowPtr pWin)
{
#if 0
  HRGN		hRgn = CreateRectRgn (0, 0, 0, 0);
  
  if (hRgn != NULL)
    {
      WalkTree (pWin->drawable.pScreen, winAddRgn, &hRgn);
      SetWindowRgn (winGetScreenPriv(pWin->drawable.pScreen)->hwndScreen,
		    hRgn, TRUE);
    }
  else
    {
      ErrorF ("winUpdateRgn fail to CreateRectRgn\n");
    }
#endif
}

#ifdef SHAPE
static
void
winReshape (WindowPtr pWin)
{
  int		nRects;
  ScreenPtr	pScreen = pWin->drawable.pScreen;
  RegionRec	rrNewShape;
  BoxPtr	pShape, pRects, pEnd;
  HRGN		hRgn, hRgnRect;
  winWindowPriv(pWin);

#if CYGDEBUG
  ErrorF ("winReshape()\n");
#endif
  
  if (pWin->parent == NULL)
    { /* If pWin is Root */
      return;
    }
  else if (pWin->parent->parent != NULL)
    { /* If pWin is not top level */
      return;
    }
  
  if (wBoundingShape(pWin)) 
    {
      REGION_INIT(pScreen, &rrNewShape, NullBox, 0);
      REGION_COPY(pScreen, &rrNewShape, wBoundingShape(pWin));
      REGION_TRANSLATE(pScreen, &rrNewShape, pWin->borderWidth,
                       pWin->borderWidth);
    }
  else
    {
      if (pWinPriv->hRgn != NULL)
        {
	  DeleteObject (pWinPriv->hRgn);
	  pWinPriv->hRgn = NULL;
	  return;
        }
    }
  
  nRects = REGION_NUM_RECTS(&rrNewShape);
  pShape = REGION_RECTS(&rrNewShape);
  
  if (nRects > 0)
    {
      hRgn = CreateRectRgn (0, 0, 0, 0);
      for (pRects = pShape, pEnd = pShape+nRects; pRects < pEnd; pRects++)
        {
	  hRgnRect = CreateRectRgn (pRects->x1, pRects->y1,
				    pRects->x2, pRects->y2);
	  if (hRgnRect == NULL)
	    {
	      ErrorF("winReshape - CreateRectRgn() failed\n");
	    }

	  if (CombineRgn (hRgn, hRgn, hRgnRect, RGN_OR) == ERROR)
	    {
	      ErrorF("winReshape - CombineRgn() failed\n");
	    }

	  DeleteObject (hRgnRect);
        }

      if (pWinPriv->hRgn != NULL)
	{
	  DeleteObject (pWinPriv->hRgn);
	  pWinPriv->hRgn = NULL;
	}
      
      pWinPriv->hRgn = hRgn;
    }
  else
    {
      if (pWinPriv->hRgn != NULL)
        {
	  DeleteObject (pWinPriv->hRgn);
	  pWinPriv->hRgn = NULL;
        }
      
    }

  REGION_UNINIT(pScreen, &rrNewShape);
  
  return;
}
#endif

LRESULT CALLBACK
winTopLevelWindowProc (HWND hwnd, UINT message, 
	       WPARAM wParam, LPARAM lParam)
{
  POINT			ptMouse;
  HDC			hdcUpdate;
  RECT			rcClient;
  PAINTSTRUCT		ps;
  WindowPtr		pWin = NULL;
  winPrivWinPtr	        pWinPriv = NULL;
  //ScreenPtr		pScreen = NULL;
  winPrivScreenPtr	pScreenPriv = NULL;
  winScreenInfo		*pScreenInfo = NULL;
  HWND			hwndScreen = NULL;
  DrawablePtr		pDraw = NULL;
  int			iScanCode;
  int			i;
  int		        iX, iY, iWidth, iHeight, iBorder;
  winWMMessageRec	wmMsg;
  
  if ((pWin = GetProp (hwnd, WIN_WINDOW_PROP)) != NULL)
    {
      pDraw		= &pWin->drawable;
      //pScreen		= pWin->drawable.pScreen;
      //ErrorF("pScreen %08X\n", pScreen);
      pWinPriv = winGetWindowPriv(pWin);
      pScreenPriv	= pWinPriv->pScreenPriv;
      pScreenInfo	= pScreenPriv->pScreenInfo;
      hwndScreen	= pScreenPriv->hwndScreen;
      wmMsg.msg		= 0;
      wmMsg.hwndWindow	= hwnd;
      wmMsg.iWindow	= (Window)GetProp(hwnd, WIN_WID_PROP);
      wmMsg.iX		= pWinPriv->iX;
      wmMsg.iY		= pWinPriv->iY;
      wmMsg.iWidth	= pWinPriv->iWidth;
      wmMsg.iHeight	= pWinPriv->iHeight;
#if 0
      ErrorF("hWnd %08X\n", hwnd);
      ErrorF("pWin %08X\n", pWin);
      ErrorF("pDraw %08X\n", pDraw);
      ErrorF("\ttype %08X\n", pWin->drawable.type);
      ErrorF("\tclass %08X\n", pWin->drawable.class);
      ErrorF("\tdepth %08X\n", pWin->drawable.depth);
      ErrorF("\tbitsPerPixel %08X\n", pWin->drawable.bitsPerPixel);
      ErrorF("\tid %08X\n", pWin->drawable.id);
      ErrorF("\tx %08X\n", pWin->drawable.x);
      ErrorF("\ty %08X\n", pWin->drawable.y);
      ErrorF("\twidth %08X\n", pWin->drawable.width);
      ErrorF("\thenght %08X\n", pWin->drawable.height);
      ErrorF("\tpScreen %08X\n", pWin->drawable.pScreen);
      ErrorF("\tserialNumber %08X\n", pWin->drawable.serialNumber);
      ErrorF("g_iWindowPrivateIndex %d\n", g_iWindowPrivateIndex);
      ErrorF("pWinPriv %08X\n", pWinPriv);
      ErrorF("pScreenPriv %08X\n", pScreenPriv);
      ErrorF("pScreenInfo %08X\n", pScreenInfo);
      ErrorF("hwndScreen %08X\n", hwndScreen);
#endif
    }

  switch (message)
    {
    case WM_CREATE:
      SetProp (hwnd, WIN_WINDOW_PROP, ((LPCREATESTRUCT) lParam)->lpCreateParams);
      SetProp (hwnd, WIN_WID_PROP, winGetWindowID(((LPCREATESTRUCT) lParam)->lpCreateParams));
      return 0;
      
    case WM_PAINT:
      ErrorF ("winXWindowProc()WM_PAINT\n");
      //SendMessage(hwndScreen, message, wParam, lParam);
      
      if(hwndScreen != NULL)
        {
	  hdcUpdate = BeginPaint (hwnd, &ps);
	  GetClientRect (hwnd, &rcClient);
	  ErrorF("ClientRect %d, %d - %d, %d\n", rcClient.left, rcClient.top,
	  	 rcClient.right, rcClient.bottom);

	  iBorder = wBorderWidth (pWin);
	  iX = pWin->drawable.x;// - iBorder;
	  iY = pWin->drawable.y;// - iBorder;
	  iWidth = pWin->drawable.width;// + iBorder * 2;
	  iHeight = pWin->drawable.height;// + iBorder * 2;

	  if(!BitBlt (hdcUpdate,
		  0, 0,
		  iWidth,
		  iHeight,
		  pScreenPriv->hdcShadow,
		  iX, iY,
/*
		  rcClient.right - rcClient.left,
		  rcClient.bottom - rcClient.top,
		  pScreenPriv->hdcShadow,
		  rcClient.left, rcClient.top,
*/
		  SRCCOPY))
	    {
		LPVOID lpMsgBuf;
		
		FormatMessage( 
		    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		    FORMAT_MESSAGE_FROM_SYSTEM | 
		    FORMAT_MESSAGE_IGNORE_INSERTS,
		    NULL,
		    GetLastError(),
		    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		    (LPTSTR) &lpMsgBuf,
		    0, NULL);

		ErrorF("BitBltFailed %s\n", (LPSTR)lpMsgBuf);
		LocalFree( lpMsgBuf );
	    }
	  EndPaint (hwndScreen, &ps);
	}
      return 0;
    case WM_MOUSEMOVE:
    case WM_NCMOUSEMOVE:
    case WM_LBUTTONDBLCLK:
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_MBUTTONDBLCLK:
    case WM_MBUTTONDOWN:
    case WM_MBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
      ErrorF ("winXWindowProc()WM_*BUTTON*\n");
      ptMouse.x = GET_X_LPARAM(lParam);
      ptMouse.y = GET_Y_LPARAM(lParam),
      ClientToScreen(hwnd, &ptMouse);
      //ScreenToClient(hwndScreen, &ptMouse);
      SendMessage(hwndScreen, message, wParam, MAKELONG(ptMouse.x, ptMouse.y));
      return  DefWindowProc (hwnd, message, wParam, lParam);
    case WM_MOUSELEAVE:
      break;
    case WM_MOUSEWHEEL:
      ErrorF ("winXWindowProc()WM_MOUSEWHEEL\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      return 0;
#if 0
    case WM_SETFOCUS:
    case WM_KILLFOCUS:
      ErrorF ("winXWindowProc()WM_*FORCUS\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      return 0;
#endif
    case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
    case WM_SYSKEYUP:
    case WM_KEYUP:
    case WM_DEADCHAR:
    case WM_SYSDEADCHAR:
      ErrorF ("winXWindowProc()WM_*KEY*\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      return 0;

    case WM_HOTKEY:
      ErrorF ("winXWindowProc()WM_HOTKEY\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      return 0;
#if 1
    case WM_ACTIVATE:
      ErrorF ("winXWindowProc()WM_ACTIVATE\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      if (LOWORD(wParam) != WA_INACTIVE)
	{
#if 0
	  PostThreadMessage (pScreenPriv->dwWMThreadID,
	  		     WM_WM_MAP, pWinPriv, 0);
	  PostThreadMessage (pScreenPriv->dwWMThreadID,
			     WM_WM_RAISE, pWinPriv, 0);
#else
	  wmMsg.msg = WM_WM_MAP;
	  winSendMessageToWM (&wmMsg);
	  wmMsg.msg = WM_WM_RAISE;
	  winSendMessageToWM (&wmMsg);
#endif
	}
      return 0;

    case WM_ACTIVATEAPP:
      ErrorF ("winXWindowProc()WM_ACTIVATEAPP\n");
      SendMessage(hwndScreen, message, wParam, lParam);
      return 0;

#endif
    case WM_CLOSE:
      ErrorF ("winXWindowProc()WM_CLOSE\n");
      if (pWinPriv->fXKilled)
        {
	  DestroyWindow (hwnd);
	}
      else
	{
	  wmMsg.msg = WM_WM_KILL;
	  winSendMessageToWM (&wmMsg);
	}
      /*PostQuitMessage(0);*/
      return 0;
    case WM_DESTROY:
      ErrorF ("winXWindowProc()WM_DESTROY\n");
      KillTimer (hwnd, s_nIDMoveEvent);
#if 1
      if (pWinPriv && (!pWinPriv->fXKilled))
	{
	  ErrorF ("winXWindowProc()WM_DESTROY - WM_WM_KILL\n");
	  wmMsg.msg = WM_WM_KILL;
	  winSendMessageToWM (&wmMsg);
	}
#endif
      ErrorF ("winXWindowProc()-WM_DESTROY\n");
      break;
    case WM_MOVE:
      ErrorF ("(%d ms)winXWindowProc()WM_MOVE\n", GetTickCount());
      pWinPriv->iX = (int)(short) LOWORD(lParam);
      pWinPriv->iY = (int)(short) HIWORD(lParam);
      ErrorF ("\t(%d, %d)\n", pWinPriv->iX, pWinPriv->iY);
#if 1
      SetTimer (hwnd, s_nIDMoveEvent, MOVE_EVENT_ELAPSE, NULL);
#else
      if (!fMoveByXServer)
	{
	  wmMsg.msg = WM_WM_MOVE;
	  wmMsg.iX = pWinPriv->iX;
	  wmMsg.iY = pWinPriv->iY;
	  winSendMessageToWM (&wmMsg);
	}
#endif
      return 0;
    case WM_SIZE:
      ErrorF ("(%d ms)winXWindowProc()WM_SIZE\n", GetTickCount());
      switch (wParam)
	{
	case SIZE_MINIMIZED:
	  //wmMsg.msg = WM_WM_MOVE;
	  //wmMsg.iWidth = pWinPriv->iWidth;
	  //wmMsg.iHeight = pWinPriv->iHeight;
	  //winSendMessageToWM (&wmMsg);
	  wmMsg.msg = WM_WM_LOWER;
	  winSendMessageToWM (&wmMsg);
	  break;
	case SIZE_RESTORED:
	case SIZE_MAXIMIZED:
	  pWinPriv->iWidth = (int)(short) LOWORD(lParam);
	  pWinPriv->iHeight = (int)(short) HIWORD(lParam);
	  ErrorF ("\t(%d, %d)\n", pWinPriv->iWidth, pWinPriv->iHeight);
	  wmMsg.msg = WM_WM_MAP;
	  wmMsg.iWidth = pWinPriv->iWidth;
	  wmMsg.iHeight = pWinPriv->iHeight;
	  winSendMessageToWM (&wmMsg);
	  SetTimer (hwnd, s_nIDMoveEvent, MOVE_EVENT_ELAPSE, NULL);
	  break;
	default:
	  break;
	}
      return 0;
    case WM_TIMER:
      ErrorF ("(%d ms)winXWindowProc()WM_TIMER\n", GetTickCount());
      if (s_nIDMoveEvent == wParam)
	{
	  KillTimer (hwnd, s_nIDMoveEvent);
	  wmMsg.msg = WM_WM_MOVE;
	  winSendMessageToWM (&wmMsg);
	}
      return 0;
    default:
      break;
    }

  return DefWindowProc (hwnd, message, wParam, lParam);
}
#if 0
{
  switch(message)
    {
    case WM_CREATE:
      ErrorF("winTopLevelWindowProc - WM_CREATE\n");
      break;
    case WM_CLOSE:
      ErrorF("winTopLevelWindowProc - WM_CLOSE\n");
      break;
    case WM_PAINT:
      ErrorF("winTopLevelWindowProc - WM_PAINT\n");
      break;
    }
  return DefWindowProc (hwnd, message, wParam, lParam);
}
#endif

static void
winCreateWindowsWindow (WindowPtr pWin)
{
  int                   iX, iY;
  int			iWidth;
  int			iHeight;
  int                   iBorder;
  HWND			hWnd;
  WNDCLASS		wc;
  ScreenPtr	        pScreen = pWin->drawable.pScreen;
  winWindowPriv(pWin);

  ErrorF ("%08X:winCreateWindowsWindow\n", pWin);

  iBorder = wBorderWidth (pWin);
  
  iX = pWin->drawable.x;// - iBorder;
  iY = pWin->drawable.y;// - iBorder;
  
  iWidth = pWin->drawable.width;// + iBorder * 2;
  iHeight = pWin->drawable.height;// + iBorder * 2;
  
  /* Setup our window class */
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = winTopLevelWindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = g_hInstance;
  wc.hIcon = LoadIcon (g_hInstance, IDI_XWIN);
  wc.hCursor = 0;
  wc.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
  wc.lpszMenuName = NULL;
  wc.lpszClassName = WINDOW_CLASS_X;
  RegisterClass (&wc);

  /* Create the window */
  hWnd = CreateWindowExA (WS_EX_TOOLWINDOW,	/* Extended styles */
			  WINDOW_CLASS_X,	/* Class name */
			  WINDOW_TITLE_X,	/* Window name */
			  WS_POPUP,
			  //WS_POPUP|WS_SYSMENU|WS_CAPTION
			  //|WS_SIZEBOX|WS_OVERLAPPEDWINDOW,
			  iX,			/* Horizontal position */
			  iY,			/* Vertical position */
			  iWidth,		/* Right edge */ 
			  iHeight,		/* Bottom edge */
			  (HWND) NULL,	/* No parent or owner window */
			  (HMENU) NULL,	/* No menu */
			  GetModuleHandle (NULL),/* Instance handle */
			  pWin);	        /* ScreenPrivates */
  if (hWnd == NULL)
    {
      ErrorF("winCreateWindowsWindow = CreateWindowExA() failed\n");
    }

  SetProp (pWinPriv->hWnd, WIN_WID_PROP, winGetWindowID(pWin));
  ErrorF("%s %d\n", WIN_WID_PROP, GetProp (pWinPriv->hWnd, WIN_WID_PROP));
  
  pWinPriv->hWnd = hWnd;
}

static void
winDestroyWindowsWindow (WindowPtr pWin)
{
  winWindowPriv(pWin);
  
  ErrorF("winDestroyWindowsWindow\n");
  if (pWinPriv->hWnd != NULL)
    {
      SetProp (pWinPriv->hWnd, WIN_WINDOW_PROP, 0);
      
      DestroyWindow (pWinPriv->hWnd);
      pWinPriv->hWnd = NULL;
      
      {
	MSG			msg;
	
	/* Process all messages on our queue */
	while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
	  {
	    if (g_hDlgDepthChange == 0 || !IsDialogMessage (g_hDlgDepthChange, &msg))
	      {
		DispatchMessage (&msg);
	      }
	  }
      }
    }
  ErrorF("-winDestroyWindowsWindow\n");
}

static void
winUpdateWindowsWindow (WindowPtr pWin)
{
  winWindowPriv(pWin);
  HWND hWnd = pWinPriv->hWnd;

  ErrorF("winUpdateWindowsWindow\n");

  if (pWin->parent != NULL && pWin->parent->parent == NULL && pWin->mapped)
    {
      if (hWnd == NULL)
	{
	  winCreateWindowsWindow (pWin);
	  assert(pWinPriv->hWnd != NULL);
	}
    }
  else
    {
      if (hWnd != NULL)
	{
	  winDestroyWindowsWindow (pWin);
	  assert(pWinPriv->hWnd == NULL);
	}
    }
  ErrorF("-winUpdateWindowsWindow\n");
}

typedef struct {
    pointer		value;
    XID			id;
} WindowIDPairRec, *WindowIDPairPtr;

static void
winFindWindow(pointer value, XID id, pointer cdata)
{
  WindowIDPairPtr wi = (WindowIDPairPtr)cdata;
  if (value == wi->value)
    {
      wi->id = id;
    }
}
static XID
winGetWindowID(WindowPtr pWin)
{
  WindowIDPairRec wi = {pWin, 0};
  ClientPtr c = wClient(pWin);
  
  FindClientResourcesByType(c, RT_WINDOW, winFindWindow, &wi);
  ErrorF("Window %d\n", wi.id);
  return wi.id;
}

